#==============================================================================
#
#  $Id: message.py,v 1.12 2000/06/25 21:37:31 mike Exp $
#
"""
   This module houses the spugmail message data object and all suporting 
   objects.
"""
#
#  Copyright (C) 1999 Michael A. Muller
#
#  Permission is granted to use, modify and redistribute this code,
#  providing that the following conditions are met:
#
#  1) This copyright/licensing notice must remain intact.
#  2) If the code is modified and redistributed, the modifications must 
#  include documentation indicating that the code has been modified.
#  3) The author(s) of this code must be indemnified and held harmless
#  against any damage resulting from the use of this code.
#
#  This code comes with ABSOLUTELY NO WARRANTEE, not even the implied 
#  warrantee of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
#
#  $Log: message.py,v $
#  Revision 1.12  2000/06/25 21:37:31  mike
#  Fixes and enhancements done while adding message forwarding.
#
#  Revision 1.11  2000/06/16 00:02:17  mike
#  Added code to process attachments in embedded message/rfc822 sections.
#
#  Revision 1.10  2000/05/25 01:26:47  mike
#  Added support for adding attachments.
#
#  Revision 1.9  2000/05/11 01:17:42  mike
#  Made parameter processing more generic, moved the code out of getContentType()
#
#  Revision 1.8  2000/05/03 23:48:52  mike
#  Added code to view content of a message/rfc822 attachment.
#
#  Revision 1.7  2000/04/30 21:06:07  mike
#  small fixes for message parsing and generation.
#
#  Revision 1.6  2000/04/17 12:35:42  mike
#  Fixed address list and setitem stuff to work with index
#
#  Revision 1.5  2000/04/16 17:45:02  mike
#  Moved the code to search for the real text body of the message into
#  a getText() method.
#
#  Revision 1.4  2000/04/14 01:23:31  mike
#  Fixed parameter parsing for content type, fixed docs on asAddressList()
#  to indicate that it can accept header indexes.
#
#  Revision 1.3  2000/04/06 00:42:30  mike
#  Since I got a "Multipart/Mixed" type from WebTV, converted content type to
#  lower case.
#
#  Revision 1.2  2000/04/02 22:40:38  mike
#  Added header manipulation methods, added the writeTo() method, which
#  writes the message in standard, RFC-822 format.
#
#  Revision 1.1.1.1  2000/04/02 19:09:03  mike
#  Spug Mailer and libraries.
#
#
#==============================================================================

import string, re, rfc822, mimetools
from types import StringType, IntType

typeRx = re.compile('([^;\s]+)\s*(;\s*)?')
parmRx = re.compile('([^=\s]+)(=([^;]*|"[^"]*"))?(;)?\s*')
headerRx = re.compile('(\S+):(.*)', re.S)

class MessageInfo:
   """
      Instances of this track all of the information about a message that
      mail clients typically want to display.  These are stored in the
      following public variables:
      
      /sender/::
         [string] Name of the sender (this is usually just the name: e-mail
	 address is stripped from it unless there is no full name, in which
	 case this field contains the e-mail address.
      /recipients/::
         [string] recipient names and e-mail addresses.
      /date/::
         [tuple] date represented as a 9-tuple suitable for use by mktime.
      /subject/::
         [string] subject line.
      /unread/::
         [int] Set to true if the message has not been read (i.e. if the
         "status" header has an "R" in it.
      /index/::
         [int] index of the message in a mailbox. [maybe this should
         be a generalized key - so that it can be used to locate a message
         independant of a storage medium].
   """
   
   def __init__(self, sender, recipients, date, subject, unread, index):
      self.sender = sender
      self.recipients = recipients
      self.date = date
      self.subject = subject
      self.unread = unread
      self.index = index

class MessageHeaders:
   """
      This class allows message headers to be referenced as either a 
      list (where a numeric index returns a tuple consisting of the
      header name and its value).  Or a dictionary (where a header name
      returns the value of the first matching header.
   """
   
   def __init__(self):
      "Constructs an empty instance."
      self.__hdrs = []
   
   def addHeader(self, header, contents):
      "Adds the header and its contents to the end of the header list."
      self.__hdrs.append(string.lower(header), contents)

   def parseFrom(self, headerList):
      """
         Parses the header information from /headerList/, a list of strings
         such as that obtained from an RFC822 Message /headers/ attribute.
      """
      curHdr = ''
      curHdrVal = ''
      for hdr in headerList:
         
         # check to see if the current line is a header line
         m = headerRx.match(hdr)
         if m:
            
            # it is.  create a new header line
            
            # store any existing header
            if curHdr:
               self.addHeader(curHdr, string.strip(curHdrVal))
            
            # record the header and its value
            curHdr = m.group(1)
            curHdrVal = m.group(2)
         else:
            
            # it is not - it must be a continuation line, so we add it to the
            # existing headers value
            curHdrVal = curHdrVal + hdr      
      
      # add the final header
      if curHdr:
         self.addHeader(curHdr, string.strip(curHdrVal))
   
   def __getitem__(self, headerOrIndex):
      """
         If /headerOrIndex/ is an integer, this returns the header and 
         contents of the header in that position.  If it is a string, returns
         the first matching contents of the header.
         
         Raises a *KeyError* or *IndexError* when appropriate.
      """
      parmType = type(headerOrIndex)
      if parmType is StringType:
         header = string.lower(headerOrIndex)
         for hdr, val in self.__hdrs:
            if hdr == header:
               return val
         else:
            raise KeyError(header)
      elif parmType is IntType:
         return self.__hdrs[headerOrIndex]
      else:
         raise TypeError()

   def __setitem__(self, headerOrIndex, newVal):
      """
         Sets the given header to the given value.
         
         If /headerOrIndex/ is an integer, /newVal/ should be a tuple
         of header and value.
      """
      if type(headerOrIndex) is StringType:
         header = string.lower(headerOrIndex)
         
         # find the header
         for i in range(len(self.__hdrs)):
            hdr, val = self.__hdrs[i]
            
            # found it: set it
            if hdr == header:
               self.__hdrs[i] = (hdr, newVal)
               break
         
         # does not exist: add it
         else:
            self.addHeader(header, newVal)
      else:
         header = string.lower(newVal[0])
         self.__hdrs[headerOrIndex] = header, newVal[1]

   def __delitem__(self, headerOrIndex):
      """
         Deletes the indicated header.
         parms:
            headerOrIndex::
               [string or int] either an index or a header name.  If a
               header name, the first instance of the header is deleted.
      """
      if type(headerOrIndex) is IntType:
         del self.__hdrs[headerOrIndex]
      else:
         for i in range(len(self.__hdrs)):
            if self.__hdrs[i][0] == headerOrIndex:
               del self.__hdrs[i]
               break
         else:
            raise KeyError(headerOrIndex)

   def __len__(self):
      "Returns the number of headers"
      return len(self.__hdrs)

   def has_key(self, header):
      "Returns true if /header/ is a defined header."
      
      header = string.lower(header)
      for hdr, val in self.__hdrs:
         if hdr == header:
            return 1
      else:
         return 0

   def asAddressList(self, headerOrIndex):
      """
         Returns the indicated header's contents converted to an addresslist.
         You must use the "addresslist" attribute of the return value to get
         at the list of tuples (canonical name, e-mail address) that 
         constitute the address entries.
         
         parms:
            headerOrIndex::
               [string or int] header name or index.
      """
      if type(headerOrIndex) is StringType:
         value = self[headerOrIndex]
      else:
         value = self[headerOrIndex][1]
      return rfc822.AddressList(value)

   def moveToEnd(self, header):
      """
         Moves the first occurance of the given header to the end of the list.
         Does nothing if the header does not exist.
         
         parms:
            header::
               [string] header name.
      """
      for i in range(len(self.__hdrs)):
         if self.__hdrs[i][0] == header:
            info = self.__hdrs[i]
            del self.__hdrs[i]
            self.__hdrs.append(info)
            break

   def parseParms(self, headerOrIndex):
      """
         Parses the "parameters" of the header.  Assumes that the
         header is of the form "type; parm=value...".
         
         If the header is not of this form, returns *None*.
         
         Returns a tuple consisting of the type and a parameter dictionary
         if successful, raises a *ValueError* if the format is incorrect.
      """
      content = self[headerOrIndex]
      m = typeRx.match(content)
      if not m:
         raise ValueError('Unknown header format')
      
      typeField = string.lower(m.group(1))
      parms = {}
      hasSemi = m.group(2)
      
      # as long as there is the possibility of matching a parameter,
      # keep looking for one.
      while hasSemi:
         content = content[m.end():]
         m = parmRx.match(content)
         if m:
            key = string.lower(m.group(1))
            
            # is there an assignment?
            if m.group(2):
               val = m.group(3)
               
               # if the value has quotes around it, strip them
               if val[0] == '"' and val[-1] == '"':
                  val = val[1:-1]
               
            else:
               val = 1
            
            parms[key] = val
            
            hasSemi = m.group(4)
         else:
            break
      
      return typeField, parms

   def __str__(self):
      """
         Returns the headers as a big formatted string (standard RFC-822
         style).
      """
      all = ''
      for hdr, val in self.__hdrs:
         all = all + '%s: %s\n' % (string.capitalize(hdr), val)
      return all
         

class Message:
   """
      Public Variables:
      
      /headers/::
         (@MessageHeaders) Set of all headers for the object.
      /body/::
         (string) body of the message.
      /attachments/::
         (list<@Message>) mime attachments.
   """
   
   def __parseMessage(self, lines):
      """
         Parses a MIME message component from a list of lines.  Returns
         a new message object.
      """
      
      # parse the headers
      headers = []
      while lines:
         line = lines[0]
         del lines[0]
         if not string.strip(line):
            break
         
         headers.append(line)
      
      headersObj = MessageHeaders()
      headersObj.parseFrom(headers)
      msg = Message(headersObj, string.join(lines, '\n'))
      return msg
   
   def __parseAttachments(self):
      "Parses and returns MIME attachments from the body of the message."
      
      attachments = []

      # check for mime-type
      contentType, parms = self.getContentType()
      if contentType[:10] == 'multipart/':
         
         # get the boundary string - add a double-dash
         # to the beginning because that appears to always be there
         try:
            boundary = '--' + parms['boundary']
         except KeyError:
            # if there was no boundary string defined, blow off treating
            # this as a mime message
            self.headers.addHeader('X-Spug-Error',
                                   "Boundary parameter not found in MIME "
                                    "message."
                                   )
            return []
         
         # define a variable for the terminating boundary
         terminator = boundary + '--'
         
         parsedBody = 0
         lines = []

         # parse the message body
         for line in string.split(self.body, '\n'):
            strippedLine = string.rstrip(line)
            if strippedLine == boundary or strippedLine == terminator:
               if not parsedBody:
                  self.body = string.join(lines)
                  parsedBody = 1
               else:
                  attachments.append(self.__parseMessage(lines))
               
               lines = []
            else:
               lines.append(line)

         if lines:
            self.body = string.join([self.body,] + lines, '\n')
      
      elif contentType == 'message/rfc822':
         embedded = self.__parseMessage(string.split(self.body, '\n'))
         attachments.append(embedded)
         self.body = ''
               
      return attachments
   
   def __init__(self, headers, body, attachments = None):
      "If /attachments/ is not specified, parses them from the body."
      self.headers = headers
      self.body = body
      if attachments is None:
         self.attachments = self.__parseAttachments()
      else:
         self.attachments = attachments

   def getContentType(self):
      """
         Returns a tuple consisting of the content type string and a 
         dictionary of parmaters and values.  If there is no content
         type field, returns 'text/plain' and an empty dictionary.
      """
      try:
         return self.headers.parseParms('content-type')
      except (KeyError, ValueError):
         return 'text/plain', {}

   def isType(self, contentType):
      """
         Returns true if the message content type is equal to that specified.
      """
      try:
         myContentType = string.lower(self.headers['content-type'])
         
         # check to see if there is a semi-colon, if so, the content type will
         # be before the semicolon
         semiPos = string.find(myContentType, ';')
         if semiPos != -1:
            myContentType = myContentType[:semiPos]
         return myContentType == string.lower(contentType)
      except KeyError:
         return 0

   def asMessageInfo(self, index = -1):
      """
         Returns /self/ converted into a @MessageInfo instance.  /index/
         should be the message index (an integer).
      """
      
      # parse the from field
      try:
         fromField = rfc822.AddressList(self.headers['from'])
         if fromField.addresslist:
            
            # use the full name, if it exists, use the email address if not.
            if fromField.addresslist[0][0]:
               fromField = fromField.addresslist[0][0]
            else:
               fromField.addresslist[0][1]
         else:
            raise KeyError()
         
      except KeyError, ex:
         fromField = 'from ??'
      
      # parse the date field
      try:
         dateField = rfc822.parsedate(self.headers['date'])
         if not dateField:
            raise KeyError()
      except KeyError:
         # we must have a date, so we create our own.
         import time
         dateField = time.localtime(time.time())
      
      # parse the other stuff
      toField = self.headers.has_key('to') and self.headers['to'] or ''
      subjectField = self.headers.has_key('subject') and \
         self.headers['subject'] or ''
      unread = self.headers.has_key('status') and \
         'R' not in self.headers['status'] or 0
      
      return MessageInfo(fromField,
                         toField,
                         dateField,
                         subjectField,
                         unread,
                         index
                         )

   def findType(self, type):
      """
         Recursively locates a MIME component of the given type.
      """
      if self.isType(type):
         return self
         
      for att in self.attachments:
         subAtt = att.findType(type)
         if subAtt:
            return subAtt
      else:
         return None

   def fixAttachments(self, toplevel = 1):
      """
         If the message has attachments, checks to make sure that there
         is a 'content-type: multipart/mixed; boundary="..."' header, and
         moves the body of the message to the first text/plain section.
         
         parms:
            toplevel::
               [int] if true, indicates that this is the toplevel message,
               so certain transformations (moving the body, setting
               the mime-version...)               
      """
      if self.attachments:
      
         # if this is the toplevel message, perform certain transformations
         if toplevel:
         
            # make sure that there is a mime version
            if not self.headers.has_key('mime-version'):
               self.headers['mime-version'] = '1.0'
         
            # make sure that the body is the MIME introduction
            if string.strip(self.body) != 'This is a MIME/Multipart message.':
               
               # create a new attachment for the body
               bodyHeaders = MessageHeaders()
               bodyHeaders['content-type'] = 'text/plain'
               bodyAttchmt = Message(bodyHeaders, self.body, [])
               self.attachments.insert(0, bodyAttchmt)
               
               # reset the body to the special intro string.
               self.body = 'This is a MIME/Multipart message.\n'
      
         # make sure that there is a content type
         contentType, parms = self.getContentType()
         if contentType != 'multipart/mixed' or not parms.has_key('boundary'):
            self.headers['content-type'] = \
               'multipart/mixed; boundary="%s"' % mimetools.choose_boundary()
         
         # recurse through nested attachments
         import main
         for att in self.attachments:
            att.fixAttachments(0)

   def expandAttachments(self, indent = ''):
      out = ''      
      for att in self.attachments:
         contentType = att.getContentType()[0]
 
         # if this has a content disposition of "inline", expand the body
         if att.headers.has_key('content-disposition') and \
            string.lower(att.headers['content-disposition']) == 'inline':
            out = out + '\n' + '-' * 78 + '\n' + att.getText() + '\n'
         else:
            out = out + '\n' + indent + contentType + \
               att.expandAttachments(indent + '   ')
         
      return out

   def getText(self):
      """
         Returns the plain text part of the message: returns the body
         of a normal message, returns the first "`text/plain`" attachment
         of a multipart message.
      """
      if self.attachments:
         textAttachment = self.findType('text/plain')
         if textAttachment:
            text = textAttachment.body
         else:
            text = self.body
      else:
         text = self.body
      
      return text
      

   def forDisplay(self, allHeaders = 0, attachmentSummary = 1):
      out = ''
      if allHeaders:
         for hdr, val in self.headers:
            out = out + '%s: %s\n' % (hdr, val)
      else:
         for hdr in ('from', 'to', 'cc', 'date', 'subject'):
            try:
               val = self.headers[hdr]
               out = out + '%s: %s\n' % (hdr, val)
            except KeyError:
               pass

      body = self.getText()      
      out = out + '\n' + body 
      
      # add an attachment summary if requested (an indented attachment tree)
      if attachmentSummary:
         out = out + self.expandAttachments()
      
      return out

   def writeTo(self, out, style = 0):
      """
         Writes the message to an output stream in the standard, e-mail type
         format.
         parms:
            out::
               [file] A file object to write to.
            style::
               [int] indicates which special character conventions to respect. 
               Possible values are:
               0::
                  No special conventions.
               1::
                  SMTP conventions - quote any line containing nothing but
                  a period.
               2::
                  Mailbox conventions - quote any line starting with "From "
      """
      
      # write all of the headers
      for hdr, val in self.headers:
         out.write('%s: %s\n' % (string.capitalize(hdr), val))
      
      # write out the body
      out.write('\n')
      if style == 0:
         out.write(self.body)
      elif style == 1:
         for line in string.split(self.body, '\n'):
            if string.rstrip(line) == '.':
               line = '> .'
            out.write(line)
            out.write('\n')
      elif style == 2:
         for line in string.split(self.body, '\n'):
            if line[:5] == 'From ':
               line = '> ' + line
            out.write(line)
            out.write('\n')
      
      # make sure there is a newline at the end of the body
      if self.body and self.body[-1] != '\n':
         out.write('\n')
      
      # now write all attachments
      if self.attachments:
      
         # determine the boundary based on the message type: the only
         # messages that should have attachments are multipart messages
         # and message/rfc822 types (where the message is treated as
         # an attachment).  Any other types get the error blurb.
         contentType, parms = self.getContentType()
         if contentType[:10] == 'multipart/':
            boundary = parms['boundary']
         elif contentType == 'message/rfc822':
            boundary = ''
         else:
            boundary = 'NO BOUNDARY'
         
         for attach in self.attachments:
            if boundary:
               out.write('--%s\n' % boundary)
            attach.writeTo(out, style)
         
         # write the terminating boundary
         if boundary:
            out.write('--%s--\n' % boundary)
         